"""
Copyright (c) 2023 CEA-List

This program and the accompanying materials are made available under the
terms of the Eclipse Public License 2.0 which is available at
http://www.eclipse.org/legal/epl-2.0.

SPDX-License-Identifier: EPL-2.0
"""

import unittest
import aidge_core
from functools import reduce
import pathlib
import os
import sys
import subprocess
import shutil
import numpy as np


def initFiller(model):
    # Initialize parameters (weights and biases)
    for node in model.get_nodes():
        if node.type() == "Producer":
            prod_op = node.get_operator()
            value = prod_op.get_output(0)
            value.set_backend("cpu")
            tuple_out = node.output(0)[0]
            # No conv in current network
            if tuple_out[0].type() == "Conv" and tuple_out[1] == 1:
                # Conv weight
                aidge_core.xavier_uniform_filler(value)
            elif tuple_out[0].type() == "Conv" and tuple_out[1] == 2:
                # Conv bias
                aidge_core.constant_filler(value, 0.01)
            elif tuple_out[0].type() == "FC" and tuple_out[1] == 1:
                # FC weight
                aidge_core.normal_filler(value)
            elif tuple_out[0].type() == "FC" and tuple_out[1] == 2:
                # FC bias
                aidge_core.constant_filler(value, 0.01)
            else:
                pass


class test_export(unittest.TestCase):
    """Test aidge export"""

    def setUp(self):
        self.EXPORT_PATH = pathlib.Path("dummy_export")

    def tearDown(self):
        pass

    def test_generate_export(self):
        # Create model

        model = aidge_core.sequential(
            [
                aidge_core.FC(
                    in_channels=32 * 32 * 3, out_channels=512, name="InputNode"
                ),
                aidge_core.ReLU(name="Relu0"),
                aidge_core.FC(in_channels=512, out_channels=256, name="FC1"),
                aidge_core.ReLU(name="Relu1"),
                aidge_core.FC(in_channels=256, out_channels=128, name="FC2"),
                aidge_core.ReLU(name="Relu2"),
                aidge_core.FC(in_channels=128, out_channels=10, name="OutputNode"),
            ]
        )

        initFiller(model)

        # Export model
        aidge_core.export(self.EXPORT_PATH, model)
        self.assertTrue(
            self.EXPORT_PATH.is_dir(), "Export folder has not been generated"
        )
        os.makedirs(self.EXPORT_PATH / "build", exist_ok=True)

        # Test compilation of export
        install_path = (
            os.path.join(sys.prefix, "lib", "libAidge")
            if "AIDGE_INSTALL" not in os.environ
            else os.environ["AIDGE_INSTALL"]
        )

        shutil.copyfile(
            pathlib.Path(__file__).parent / "static/main.cpp",
            self.EXPORT_PATH / "main.cpp",
        )

        subprocess.check_call(
            [
                "cmake",
                str(self.EXPORT_PATH.absolute()),
                "-DPYBIND=1",
                f"-DCMAKE_INSTALL_PREFIX:PATH={install_path}",
            ],
            cwd=str(self.EXPORT_PATH / "build"),
        )
        subprocess.check_call(
            ["cmake", "--build", "."],
            cwd=str(self.EXPORT_PATH / "build"),
        )
        subprocess.check_call(
            ["cmake", "--install", "."],
            cwd=str(self.EXPORT_PATH / "build"),
        )


if __name__ == "__main__":
    unittest.main()
