/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_DATA_H_
#define AIDGE_DATA_H_

#include <cstdint>
#include <fmt/format.h>
#include <string>
#include <tuple>
#include <array>

#include "aidge/data/half.hpp"
#include "aidge/utils/Attributes.hpp"

namespace Aidge {
enum class DataType {
    Float64,
    Float32,
    Float16,
    BFloat16,
    Binary,
    Ternary,
    Int2,
    Int3,
    Int4,
    Int5,
    Int6,
    Int7,
    Int8,
    Int16,
    Int32,
    Int64,
    UInt2,
    UInt3,
    UInt4,
    UInt5,
    UInt6,
    UInt7,
    UInt8,
    UInt16,
    UInt32,
    UInt64
};

enum class DataFormat {
    Default,
    NCHW,
    NHWC,
    CHWN,
    NCDHW,
    NDHWC,
    CDHWN
};

using DataFormatTranspose = std::array<size_t, 5>;
// Permutation arrays dict to obtain DataFormat (same order as DataFormat enum)
constexpr std::array<DataFormatTranspose, 7> DataFormatTransposeDict = {{
    // Important: in this array only, dimension index must start at 1, not 0!
    // (0 is the default value)
    {},
    {1, 2, 3, 4},
    {1, 3, 4, 2},
    {2, 3, 4, 1},
    {1, 2, 3, 4, 5},
    {1, 3, 4, 5, 2},
    {2, 3, 4, 5, 1}
}};

/**
 * Get the DataFormatTranspose array to transpose data from src to dst DataFormat.
 * @param src Source DataFormat
 * @param dst Destinatin DataFormat
 * @return DataFormatTranspose Permutation array to achieve a transposition
 *         from src to dst DataFormat.
*/
constexpr inline DataFormatTranspose getDataFormatTranspose(const DataFormat& src, const DataFormat& dst) {
    // Permutation array from default format to src format
    const auto srcDefToFormat = DataFormatTransposeDict[static_cast<int>(src)];
    // Permutation array from default format to dst format
    const auto dstDefToFormat = DataFormatTransposeDict[static_cast<int>(dst)];
    // Compute permutation array from src format to default format:
    DataFormatTranspose srcFormatToDef{};
    for (size_t i = 0; i < srcDefToFormat.size(); ++i) {
        if (srcDefToFormat[i] > 0) {
            srcFormatToDef[srcDefToFormat[i] - 1] = i;
        }
        else {
            srcFormatToDef[i] = i;
        }
    }

    // Compute permutation array from src format to dst format:
    DataFormatTranspose srcToDst{};
    for (size_t i = 0; i < dstDefToFormat.size(); ++i) {
        if (dstDefToFormat[srcFormatToDef[i]] > 0) {
            srcToDst[i] = dstDefToFormat[srcFormatToDef[i]] - 1;
        }
        else {
            srcToDst[i] = i;
        }
    }

    return srcToDst;
}

class Data {
public:
    Data(const std::string& type): mType(type) {};
    constexpr const std::string& type() const {
        return mType;
    }
    virtual ~Data() = default;
    virtual std::string toString() const = 0;

private:
    const std::string mType;
};
}

namespace {
template <typename T> struct NativeType { static const Aidge::DataType type; };
template <> const Aidge::DataType NativeType<double>::type = Aidge::DataType::Float64;
template <> const Aidge::DataType NativeType<float>::type = Aidge::DataType::Float32;
template <> const Aidge::DataType NativeType<half_float::half>::type = Aidge::DataType::Float16;
template <> const Aidge::DataType NativeType<std::int8_t>::type = Aidge::DataType::Int8;
template <> const Aidge::DataType NativeType<std::int16_t>::type = Aidge::DataType::Int16;
template <> const Aidge::DataType NativeType<std::int32_t>::type = Aidge::DataType::Int32;
template <> const Aidge::DataType NativeType<std::int64_t>::type = Aidge::DataType::Int64;
template <> const Aidge::DataType NativeType<std::uint8_t>::type = Aidge::DataType::UInt8;
template <> const Aidge::DataType NativeType<std::uint16_t>::type = Aidge::DataType::UInt16;
template <> const Aidge::DataType NativeType<std::uint32_t>::type = Aidge::DataType::UInt32;
template <> const Aidge::DataType NativeType<std::uint64_t>::type = Aidge::DataType::UInt64;

template <>
const char* const EnumStrings<Aidge::DataType>::data[]
    = {"Float64", "Float32", "Float16", "BFloat16", "Binary", "Ternary",
       "Int2", "Int3", "Int4", "Int5", "Int6", "Int7", "Int8", "Int16",
       "Int32", "Int64", "UInt2", "UInt3", "UInt4", "UInt5", "UInt6",
       "UInt7", "UInt8", "UInt16", "UInt32", "UInt64"};

template <>
const char* const EnumStrings<Aidge::DataFormat>::data[]
    = {"Default", "NCHW", "NHWC", "CHWN", "NCDHW", "NDHWC", "CDHWN"};

template <Aidge::DataType D> struct cpptype {
    using type = void; // Placeholder
};
template <> struct cpptype<Aidge::DataType::Float16> { using type = half_float::half; };
template <> struct cpptype<Aidge::DataType::Float32> { using type = float; };
template <> struct cpptype<Aidge::DataType::Float64> { using type = double; };
template <> struct cpptype<Aidge::DataType::Int8> { using type = std::int8_t; };
template <> struct cpptype<Aidge::DataType::Int16> { using type = std::int16_t; };
template <> struct cpptype<Aidge::DataType::Int32> { using type = std::int32_t; };
template <> struct cpptype<Aidge::DataType::Int64> { using type = std::int64_t; };
template <> struct cpptype<Aidge::DataType::UInt8> { using type = std::uint8_t; };
template <> struct cpptype<Aidge::DataType::UInt16> { using type = std::uint16_t; };
template <> struct cpptype<Aidge::DataType::UInt32> { using type = std::uint32_t; };
template <> struct cpptype<Aidge::DataType::UInt64> { using type = std::uint64_t; };

template <Aidge::DataType D> using cpptype_t = typename cpptype<D>::type;
}


namespace Aidge {
inline auto format_as(DataType dt) { return EnumStrings<Aidge::DataType>::data[static_cast<int>(dt)]; }
inline auto format_as(DataFormat df) { return EnumStrings<Aidge::DataFormat>::data[static_cast<int>(df)]; }
}

#endif /* AIDGE_DATA_H_ */
