/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CORE_OPERATOR_FOLD_H_
#define AIDGE_CORE_OPERATOR_FOLD_H_

#include <array>
#include <cmath>    // std::floor
#include <cstddef>  // std::size_t
#include <string>
#include <utility>  // std::pair
#include <vector>

#include "aidge/data/Tensor.hpp"
#include "aidge/graph/Node.hpp"
#include "aidge/operator/OperatorTensor.hpp"
#include "aidge/operator/Producer.hpp"
#include "aidge/utils/ArrayHelpers.hpp"
#include "aidge/utils/ErrorHandling.hpp"
#include "aidge/utils/Registrar.hpp" // SET_IMPL_MACRO
#include "aidge/utils/StaticAttributes.hpp"
#include "aidge/utils/Types.h"

namespace Aidge {
enum class FoldAttr { OutputDims, StrideDims, DilationDims, KernelDims };

template <DimIdx_t DIM>
class Fold_Op : public OperatorTensor,
                public Registrable<Fold_Op<DIM>, std::string, std::shared_ptr<OperatorImpl>(const Fold_Op<DIM> &)> {

public:
    static const std::string Type;

private:
    using Attributes_ = StaticAttributes<FoldAttr,
                                        std::array<DimSize_t, DIM>,
                                        std::array<DimSize_t, DIM>,
                                        std::array<DimSize_t, DIM>,
                                        std::array<DimSize_t, DIM>>;
    template <FoldAttr e> using attr = typename Attributes_::template attr<e>;
    const std::shared_ptr<Attributes_> mAttributes;

public:
    Fold_Op() = delete;

    constexpr Fold_Op(const std::array<DimSize_t, DIM> &outputDims,
                      const std::array<DimSize_t, DIM> &kernelDims,
                      const std::array<DimSize_t, DIM> &strideDims = create_array<DimSize_t,DIM>(1),
                      const std::array<DimSize_t, DIM> &dilationDims = create_array<DimSize_t,DIM>(1))
        : OperatorTensor(Type, {InputCategory::Data}, 1),
          mAttributes(std::make_shared<Attributes_>(
            attr<FoldAttr::OutputDims>(outputDims),
            attr<FoldAttr::StrideDims>(strideDims),
            attr<FoldAttr::DilationDims>(dilationDims),
            attr<FoldAttr::KernelDims>(kernelDims))) {}

    /**
     * @brief Copy-constructor. Copy the operator attributes and its output tensor(s), but not its
     * input tensors (the new operator has no input associated).
     * @param op Operator to copy.
     */
    Fold_Op(const Fold_Op<DIM> &op)
        : OperatorTensor(op),
          mAttributes(op.mAttributes)
    {
        if (!op.backend().empty()) {
            SET_IMPL_MACRO(Fold_Op<DIM>, *this, op.backend());
        }
        else {
            mImpl = nullptr;
        }
    }

    /**
     * @brief Clone the operator using its copy-constructor.
     * @see Operator::Fold_Op
     */
    std::shared_ptr<Operator> clone() const override {
        return std::make_shared<Fold_Op<DIM>>(*this);
    }

    bool forwardDims(bool /*allowDataDependency*/ = false) override final;

    void setBackend(const std::string &name, DeviceIdx_t device = 0) override;

    inline std::shared_ptr<Attributes> attributes() const override { return mAttributes; }
    inline std::array<DimSize_t, DIM>& outputDims() const { return mAttributes->template getAttr<FoldAttr::OutputDims>(); }
    inline std::array<DimSize_t, DIM>& strideDims() const { return mAttributes->template getAttr<FoldAttr::StrideDims>(); }
    inline std::array<DimSize_t, DIM>& dilationDims() const { return mAttributes->template getAttr<FoldAttr::DilationDims>(); }
    inline std::array<DimSize_t, DIM>& kernelDims() const { return mAttributes->template getAttr<FoldAttr::KernelDims>(); }

    static const std::vector<std::string> getInputsName(){
        return {"data_input"};
    }
    static const std::vector<std::string> getOutputsName(){
        return {"data_output"};
    }
};

template <std::array<DimSize_t, 1>::size_type DIM>
inline std::shared_ptr<Node> Fold(const std::array<DimSize_t, DIM> &outputDims,
                                  const std::array<DimSize_t, DIM> &kernelDims,
                                  const std::string& name = "",
                                  const std::array<DimSize_t, DIM> &strideDims = create_array<DimSize_t,DIM>(1),
                                  const std::array<DimSize_t, DIM> &dilationDims = create_array<DimSize_t,DIM>(1)) {
    // FIXME: properly handle default w&b initialization in every cases
    static_assert(DIM<=MaxDim,"Too many kernel dimensions required by Fold, not supported");
    return std::make_shared<Node>(std::make_shared<Fold_Op<static_cast<DimIdx_t>(DIM)>>(outputDims, kernelDims, strideDims, dilationDims), name);
}

template <DimSize_t DIM>
inline std::shared_ptr<Node> Fold(
    DimSize_t const (&outputDims)[DIM],
    DimSize_t const (&kernelDims)[DIM],
    const std::string& name = "",
    const std::array<DimSize_t, DIM> &strideDims = create_array<DimSize_t,DIM>(1),
    const std::array<DimSize_t, DIM> &dilationDims = create_array<DimSize_t,DIM>(1)) {
    static_assert(DIM<=MaxDim,"Too many kernel dimensions required by Fold, not supported");
    return Fold(to_array(outputDims), to_array(kernelDims), name, strideDims, dilationDims);
}
}  // namespace Aidge

extern template class Aidge::Fold_Op<2>;

namespace {
template <>
const char *const EnumStrings<Aidge::FoldAttr>::data[] = {
    "OutputDims",
    "StrideDims",
    "DilationDims",
    "KernelDims"
};
}

#endif /* AIDGE_CORE_OPERATOR_FOLD_H_ */
