/********************************************************************************
 * Copyright (c) 2024 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CORE_OPERATOR_SHAPE_H_
#define AIDGE_CORE_OPERATOR_SHAPE_H_

#include <cstdint>  // std::int64_t
#include <memory>
#include <string>
#include <vector>

#include "aidge/backend/OperatorImpl.hpp"
#include "aidge/graph/Node.hpp"
#include "aidge/operator/OperatorTensor.hpp"
#include "aidge/utils/Registrar.hpp"
#include "aidge/utils/StaticAttributes.hpp"
#include "aidge/utils/Types.h"

namespace Aidge {
class Shape_OpImpl : public OperatorImpl {
public:
    Shape_OpImpl(const Operator& op, const std::string& backend = ""): OperatorImpl(op, backend) {}
    void forward() override;
};

enum class ShapeAttr { Start, End };

class Shape_Op : public OperatorTensor,
                public Registrable<Shape_Op,
                                   std::string,
                                   std::shared_ptr<OperatorImpl>(const Shape_Op&)> {

public:
    static const std::string Type;

private:
    using Attributes_ = StaticAttributes<ShapeAttr, std::int64_t, std::int64_t>;
    template <ShapeAttr e> using attr = typename Attributes_::template attr<e>;
    const std::shared_ptr<Attributes_> mAttributes;

public:
    Shape_Op() = delete;

    Shape_Op(const std::int64_t start, const std::int64_t end)
        : OperatorTensor(Type, {InputCategory::Data}, 1),
          mAttributes(std::make_shared<Attributes_>(
            attr<ShapeAttr::Start>(start),
            attr<ShapeAttr::End>(end)))
    {
        mImpl = std::make_shared<Shape_OpImpl>(*this);
    }

    /**
     * @brief Copy-constructor. Copy the operator attributes and its output tensor(s), but not its input tensors (the new operator has no input associated).
     * @param op Operator to copy.
     */
    Shape_Op(const Shape_Op& op)
        : OperatorTensor(op),
          mAttributes(op.mAttributes)
    {
        if (!op.backend().empty()) {
            SET_IMPL_MACRO(Shape_Op, *this, op.backend());
        }
        else {
            mImpl = std::make_shared<Shape_OpImpl>(*this);
        }
    }

    /**
     * @brief Clone the operator using its copy-constructor.
     * @see Operator::Shape_Op
     */
    std::shared_ptr<Operator> clone() const override {
        return std::make_shared<Shape_Op>(*this);
    }

    bool forwardDims(bool /*allowDataDependency*/ = false) override final;

    void setBackend(const std::string& name, DeviceIdx_t device = 0) override;

    inline std::shared_ptr<Attributes> attributes() const override { return mAttributes; }
    inline std::int64_t& start() const noexcept { return mAttributes -> getAttr<ShapeAttr::Start>(); }
    inline std::int64_t& end() const noexcept { return mAttributes -> getAttr<ShapeAttr::End>(); }

    static const std::vector<std::string> getInputsName(){
        return {"data_input"};
    }
    static const std::vector<std::string> getOutputsName(){
        return {"data_output"};
    }
};

inline std::shared_ptr<Node> Shape(const std::int64_t start = 0, const std::int64_t end = -1, const std::string& name = "") {
    return std::make_shared<Node>(std::make_shared<Shape_Op>(start, end), name);
}
} // namespace Aidge

namespace {
template <>
const char *const EnumStrings<Aidge::ShapeAttr>::data[] = {"Start", "End"};
}

#endif /* AIDGE_CORE_OPERATOR_SHAPE_H_ */
