/********************************************************************************
 * Copyright (c) 2024 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CORE_OPERATOR_SPLIT_H_
#define AIDGE_CORE_OPERATOR_SPLIT_H_

#include <memory>
#include <vector>

#include "aidge/backend/OperatorImpl.hpp"
#include "aidge/data/Tensor.hpp"
#include "aidge/graph/Node.hpp"
#include "aidge/operator/OperatorTensor.hpp"
#include "aidge/utils/Registrar.hpp"
#include "aidge/utils/StaticAttributes.hpp"
#include "aidge/utils/Types.h"

namespace Aidge {
class Split_OpImpl : public OperatorImpl {
public:
    Split_OpImpl(const Operator& op, const std::string& backend = ""): OperatorImpl(op, backend) {}
    void forward() override;
};

enum class SplitAttr { Axis, Split };

class Split_Op
    : public OperatorTensor,
      public Registrable<Split_Op, std::string, std::shared_ptr<OperatorImpl>(const Split_Op &)>,
      public StaticAttributes<SplitAttr, std::int8_t, std::vector<DimSize_t>> {

public:
    static const std::string Type;

    Split_Op() = delete;

    using Attributes_ = StaticAttributes<SplitAttr,  std::int8_t, std::vector<DimSize_t>>;
    template <SplitAttr e> using attr = typename Attributes_::template attr<e>;
    Split_Op( std::int8_t axis, DimSize_t nbOutputs, const std::vector<DimSize_t>& split)
        : OperatorTensor(Type, {InputCategory::Data, InputCategory::OptionalData}, nbOutputs),
          Attributes_(attr<SplitAttr::Axis>(axis),
                      attr<SplitAttr::Split>(split))
    {
        mImpl = std::make_shared<Split_OpImpl>(*this);
    }


    /**
     * @brief Copy-constructor. Copy the operator attributes and its output tensor(s), but not its
     * input tensors (the new operator has no input associated).
     * @param op Operator to copy.
     */
    Split_Op(const Split_Op &op)
        : OperatorTensor(op),
          Attributes_(op)
    {
        if (!op.backend().empty()) {
            SET_IMPL_MACRO(Split_Op, *this, op.backend());
        }
        else {
            mImpl = std::make_shared<Split_OpImpl>(*this);
        }
    }
public:
    /**
     * @brief Clone the operator using its copy-constructor.
     * @see Operator::Split_Op
     */
    std::shared_ptr<Operator> clone() const override { return std::make_shared<Split_Op>(*this); }

    bool dimsForwarded() const override final;
    bool forwardDims(bool allowDataDependency = false) override final;

    void setBackend(const std::string &name, DeviceIdx_t device = 0) override;

    static const std::vector<std::string> getInputsName(){
        return {"data_input", "split"};
    }
    static const std::vector<std::string> getOutputsName(){
        return {"data_output_0", "data_output_n"};
    }

};

/**
 * @brief Exract a sub-Tensor from a bigger original Tensor.
 * @param name Name of the Operator.
 * @return std::shared_ptr<Node> A Node containing the Operator.
 */
inline std::shared_ptr<Node> Split(DimSize_t nbOutput,
                                   std::int8_t axis = 0,
                                   const std::vector<DimSize_t>& split = {},
                                   const std::string &name = "") {
    return std::make_shared<Node>(std::make_shared<Split_Op>(axis, nbOutput, split), name);
}
}  // namespace Aidge

namespace {
template <>
const char *const EnumStrings<Aidge::SplitAttr>::data[] = { "Axis", "Split" };
}

#endif /* AIDGE_CORE_OPERATOR_SPLIT_H_ */
