/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CORE_OPERATOR_TRANSPOSE_H_
#define AIDGE_CORE_OPERATOR_TRANSPOSE_H_

#include <array>
#include <cmath>
#include <numeric>
#include <vector>

#include "aidge/graph/Node.hpp"
#include "aidge/operator/OperatorTensor.hpp"
#include "aidge/operator/Producer.hpp"
#include "aidge/utils/StaticAttributes.hpp"
#include "aidge/utils/Registrar.hpp"
#include "aidge/utils/Types.h"

namespace Aidge {
class TransposeImpl : public OperatorImpl {
public:
    TransposeImpl(const Operator& op, const std::string& backend = "")
        : OperatorImpl(op, backend)
    {}
    void forward() override;
};

enum class TransposeAttr { OutputDimsOrder };

class Transpose_Op : public OperatorTensor,
                public Registrable<Transpose_Op, std::string, std::shared_ptr<OperatorImpl>(const Transpose_Op&)> {

public:
    static const std::string Type;


private:
    using Attributes_ = StaticAttributes<TransposeAttr, std::vector<DimSize_t>>;
    template <TransposeAttr e> using attr = typename Attributes_::template attr<e>;
    const std::shared_ptr<Attributes_> mAttributes;

public:
    Transpose_Op() = delete;

    Transpose_Op(const std::vector<DimSize_t> &outputDimsOrder)
        : OperatorTensor(Type, {InputCategory::Data}, 1),
          mAttributes(std::make_shared<Attributes_>(
            attr<TransposeAttr::OutputDimsOrder>(outputDimsOrder)))
    {
        mImpl = std::make_shared<TransposeImpl>(*this);
    }

    /**
     * @brief Copy-constructor. Copy the operator attributes and its output tensor(s), but not its input tensors (the new operator has no input associated).
     * @param op Operator to copy.
     */
    Transpose_Op(const Transpose_Op& op)
        : OperatorTensor(op),
          mAttributes(op.mAttributes)
    {
        if (!op.backend().empty()) {
            SET_IMPL_MACRO(Transpose_Op, *this, op.backend());
        }
        else {
            mImpl = std::make_shared<TransposeImpl>(*this);
        }
    }

    /**
     * @brief Clone the operator using its copy-constructor.
     * @see Operator::Transpose_Op
     */
    std::shared_ptr<Operator> clone() const override {
        return std::make_shared<Transpose_Op>(*this);
    }

    bool forwardDims(bool /*allowDataDependency*/ = false) override final;

    void setBackend(const std::string &name, DeviceIdx_t device = 0) override;

    inline std::shared_ptr<Attributes> attributes() const override { return mAttributes; }
    inline std::vector<DimSize_t>& outputDimsOrder() const noexcept { return mAttributes -> getAttr<TransposeAttr::OutputDimsOrder>(); }

    static const std::vector<std::string> getInputsName(){
        return {"data_input"};
    }
    static const std::vector<std::string> getOutputsName(){
        return {"data_output"};
    }
};

inline std::shared_ptr<Node> Transpose(const std::vector<DimSize_t> &outputDimsOrder,
                                           const std::string& name = "") {
    return std::make_shared<Node>(std::make_shared<Transpose_Op>(outputDimsOrder), name);
}
}  // namespace Aidge

namespace {
template <>
const char *const EnumStrings<Aidge::TransposeAttr>::data[] = {"OutputDimsOrder"};
}

#endif /* AIDGE_CORE_OPERATOR_TRANSPOSE_H_ */
