/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CORE_OPERATOR_UNFOLD_H_
#define AIDGE_CORE_OPERATOR_UNFOLD_H_

#include <array>
#include <cmath>    // std::floor
#include <cstddef>  // std::size_t
#include <string>
#include <utility>  // std::pair
#include <vector>

#include "aidge/data/Tensor.hpp"
#include "aidge/graph/Node.hpp"
#include "aidge/operator/OperatorTensor.hpp"
#include "aidge/operator/Producer.hpp"
#include "aidge/utils/ArrayHelpers.hpp"
#include "aidge/utils/ErrorHandling.hpp"
#include "aidge/utils/Registrar.hpp" // SET_IMPL_MACRO
#include "aidge/utils/StaticAttributes.hpp"
#include "aidge/utils/Types.h"

namespace Aidge {
template <DimIdx_t DIM>
class Unfold_OpImpl : public OperatorImpl {
public:
    Unfold_OpImpl(const Operator& op, const std::string& backend = ""): OperatorImpl(op, backend) {}
    void forward() override;
};

enum class UnfoldAttr { StrideDims, DilationDims, KernelDims };

template <DimIdx_t DIM>
class Unfold_Op : public OperatorTensor,
                public Registrable<Unfold_Op<DIM>, std::string, std::shared_ptr<OperatorImpl>(const Unfold_Op<DIM> &)> {

public:
    static const std::string Type;

private:
    using Attributes_ = StaticAttributes<UnfoldAttr,
                                        std::array<DimSize_t, DIM>,
                                        std::array<DimSize_t, DIM>,
                                        std::array<DimSize_t, DIM>>;
    template <UnfoldAttr e> using attr = typename Attributes_::template attr<e>;
    const std::shared_ptr<Attributes_> mAttributes;

public:
    Unfold_Op() = delete;

    constexpr Unfold_Op(const std::array<DimSize_t, DIM> &kernelDims,
                      const std::array<DimSize_t, DIM> &strideDims = create_array<DimSize_t,DIM>(1),
                      const std::array<DimSize_t, DIM> &dilationDims = create_array<DimSize_t,DIM>(1))
        : OperatorTensor(Type, {InputCategory::Data}, 1),
          mAttributes(std::make_shared<Attributes_>(
            attr<UnfoldAttr::StrideDims>(strideDims),
            attr<UnfoldAttr::DilationDims>(dilationDims),
            attr<UnfoldAttr::KernelDims>(kernelDims)))
    {
        mImpl = std::make_shared<Unfold_OpImpl<DIM>>(*this);  
    }

    /**
     * @brief Copy-constructor. Copy the operator attributes and its output tensor(s), but not its
     * input tensors (the new operator has no input associated).
     * @param op Operator to copy.
     */
    Unfold_Op(const Unfold_Op<DIM> &op)
        : OperatorTensor(op),
          mAttributes(op.mAttributes)
    {
        if (!op.backend().empty()) {
            SET_IMPL_MACRO(Unfold_Op<DIM>, *this, op.backend());
        }
        else {
            mImpl = std::make_shared<Unfold_OpImpl<DIM>>(*this);  
        }
    }

    /**
     * @brief Clone the operator using its copy-constructor.
     * @see Operator::Unfold_Op
     */
    std::shared_ptr<Operator> clone() const override {
        return std::make_shared<Unfold_Op>(*this);
    }

    bool forwardDims(bool /*allowDataDependency*/ = false) override final;

    void setBackend(const std::string &name, DeviceIdx_t device = 0) override;

    inline std::shared_ptr<Attributes> attributes() const override { return mAttributes; }
    inline std::array<DimSize_t, DIM>& strideDims() const { return mAttributes->template getAttr<UnfoldAttr::StrideDims>(); }
    inline std::array<DimSize_t, DIM>& dilationDims() const { return mAttributes->template getAttr<UnfoldAttr::DilationDims>(); }
    inline std::array<DimSize_t, DIM>& kernelDims() const { return mAttributes->template getAttr<UnfoldAttr::KernelDims>(); }

    static const std::vector<std::string> getInputsName(){
        return {"data_input"};
    }
    static const std::vector<std::string> getOutputsName(){
        return {"data_output"};
    }
};

template <std::array<DimSize_t, 1>::size_type DIM>
inline std::shared_ptr<Node> Unfold(const std::array<DimSize_t, DIM> &kernelDims,
                                  const std::string& name = "",
                                  const std::array<DimSize_t, DIM> &strideDims = create_array<DimSize_t,DIM>(1),
                                  const std::array<DimSize_t, DIM> &dilationDims = create_array<DimSize_t,DIM>(1)) {
    // FIXME: properly handle default w&b initialization in every cases
    static_assert(DIM<=MaxDim,"Too many kernel dimensions required by Unfold, not supported");
    return std::make_shared<Node>(std::make_shared<Unfold_Op<static_cast<DimIdx_t>(DIM)>>(kernelDims, strideDims, dilationDims), name);
}

template <DimSize_t DIM>
inline std::shared_ptr<Node> Unfold(
    DimSize_t const (&kernelDims)[DIM],
    const std::string& name = "",
    const std::array<DimSize_t, DIM> &strideDims = create_array<DimSize_t,DIM>(1),
    const std::array<DimSize_t, DIM> &dilationDims = create_array<DimSize_t,DIM>(1)) {
    static_assert(DIM<=MaxDim,"Too many kernel dimensions required by Unfold, not supported");
    return Unfold(to_array(kernelDims), name, strideDims, dilationDims);
}
}  // namespace Aidge

extern template class Aidge::Unfold_Op<2>;

namespace {
template <>
const char *const EnumStrings<Aidge::UnfoldAttr>::data[] = {
    "StrideDims",
    "DilationDims",
    "KernelDims"
};
}

#endif /* AIDGE_CORE_OPERATOR_UNFOLD_H_ */
