/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#include "aidge/operator/Fold.hpp"

#include <cmath>      // std::floor
#include <cstddef>    // std::size_t
#include <stdexcept>  // std::runtime_error
#include <string>
#include <utility>    // std::pair
#include <vector>

#include "aidge/data/Tensor.hpp"
#include "aidge/utils/ErrorHandling.hpp"
#include "aidge/utils/Registrar.hpp"
#include "aidge/utils/Types.h"

template <Aidge::DimIdx_t DIM>
const std::string Aidge::Fold_Op<DIM>::Type = "Fold";

template <Aidge::DimIdx_t DIM>
bool Aidge::Fold_Op<DIM>::forwardDims(bool /*allowDataDependency*/) {
    if (inputsAssociated()) {
        auto dims(getInput(0)->dims());
        DimSize_t k = 1;
        DimSize_t l = 1;

        for (std::size_t dim = 0; dim < this->kernelDims().size() ; ++dim) {
            const DimSize_t kernelExtent = this->dilationDims()[dim] *
                                                    (this->kernelDims()[dim] - 1) + 1;

            k *= this->kernelDims()[dim];
            l *= 1 + static_cast<DimSize_t>(
                    floor(static_cast<float>(this->outputDims()[dim] - kernelExtent) /
                            static_cast<float>(this->strideDims()[dim])));
        }

        AIDGE_ASSERT(dims[dims.size() - 2] % k == 0 , "Fold: input number of channels ({}) is not divisible by the product of provided kernel dims ({})!",
            dims[dims.size() - 2], k);
        AIDGE_ASSERT(dims[dims.size() - 1] == l, "Fold: mismatch between expected input 3rd dim {} and provided input 3rd dim {}",
            dims[dims.size() - 1], l);

        dims[dims.size() - 2] /= k;
        dims.pop_back();
        dims.insert(dims.end(), this->outputDims().begin(), this->outputDims().end());
        mOutputs[0]->resize(dims);
        return true;
    }

    return false;
}

template <Aidge::DimIdx_t DIM>
void Aidge::Fold_Op<DIM>::setBackend(const std::string &name, Aidge::DeviceIdx_t device) {
    SET_IMPL_MACRO(Fold_Op<DIM>, *this, name);
    mOutputs[0]->setBackend(name, device);
}

template class Aidge::Fold_Op<2>;