/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#include "aidge/operator/Resize.hpp"

#include <cstddef>    // std::size_t
#include <cstdint>    // std::int64_t
#include <stdexcept>  // std::runtime_error
#include <string>
#include <vector>
#include <fmt/core.h>

#include "aidge/backend/OperatorImpl.hpp"
#include "aidge/data/Tensor.hpp"
#include "aidge/utils/ErrorHandling.hpp"
#include "aidge/utils/Types.h"

const std::string Aidge::Resize_Op::Type = "Resize";

bool Aidge::Resize_Op::dimsForwarded() const {
    // in case of ROI add getInput(1) condition
    if ((getInput(1) && !getInput(1)->empty())
        || (getInput(2) && !getInput(2)->empty())
        || (getInput(3) && !getInput(3)->empty())
        )
    {
        // output dims are data dependent
        return false;
    }

    return OperatorTensor::dimsForwarded();
}

bool Aidge::Resize_Op::forwardDims(bool allowDataDependency) {
    if (inputsAssociated()) {
        AIDGE_ASSERT(getInput(0)->nbDims() == 4,
            "input tensor must have dimensions = 4 (batch, channel, height, width).");

        const bool input1ROIPresent           = getInput(1) && !getInput(1)->empty();
        const bool input2ScalesPresent        = getInput(2) && !getInput(2)->empty();
        const bool input3SizesPresent         = getInput(3) && !getInput(3)->empty();

        AIDGE_ASSERT(input2ScalesPresent != input3SizesPresent, "Only one of scales and  sizes can be specified.")

        if (input1ROIPresent) {
            AIDGE_THROW_OR_ABORT(std::runtime_error, "Input #1 (ROI) is given and it is not supported.");
        }
        else if (input2ScalesPresent)  {
            if (!allowDataDependency) {
                Log::warn("Resize_Op: cannot execute forwardDims() as the output dimensions depend on the input #2");
                return false;
            }

            AIDGE_ASSERT(getInput(0)->nbDims() ==  getInput(2)->size(),
                "input #0 and input #2 (Scales) must have the same dimensions.");

            std::vector<DimSize_t>      outDims = getInput(0)->dims();
            const std::vector<DimSize_t> inDims = getInput(0)->dims();

            std::shared_ptr<Tensor> fallback;
            const auto& scales = getInput(2)->refCastFrom(fallback, NativeType<int64_t>::type, "cpu");

            for (std::size_t dim=0; dim < getInput(2)->size(); ++dim) {
                outDims[dim] = inDims[dim]*static_cast<int64_t*>(scales.getImpl()->hostPtr())[dim];
            }

            mOutputs[0]->resize(outDims);
            return true;
        }
        else if (input3SizesPresent) {
            if (!allowDataDependency) {
                Log::warn("Resize_Op: cannot execute forwardDims() as the output dimensions depend on the input #3");
                return false;
            }

            AIDGE_ASSERT(getInput(0)->nbDims() ==  getInput(3)->size(),
                "input #0 and input #3 (Sizes) must have the same dimensions.");

            std::vector<DimSize_t> outDims = getInput(0)->dims();

            std::shared_ptr<Tensor> fallback;
            const auto& sizes = getInput(3)->refCastFrom(fallback, NativeType<int64_t>::type, "cpu");

            for (std::size_t dim=0; dim < getInput(3)->size(); ++dim) {            
                outDims[dim] = static_cast<int64_t*>(sizes.getImpl()->hostPtr())[dim];
            }
            
            mOutputs[0]->resize(outDims);
            return true;
        }
        else {
            AIDGE_THROW_OR_ABORT(std::runtime_error, "Error: Either Input #2 or Input #3 must be present.");
        }
    }

    return false; 
}

void Aidge::Resize_Op::setBackend(const std::string& name, Aidge::DeviceIdx_t device) {
    SET_IMPL_MACRO(Resize_Op, *this, name);
    mOutputs[0]->setBackend(name, device);

    // By default, automatically set backend for all inputs: roi, scales and sizes 
    if(getInput(1)) {
        getInput(1)->setBackend(name, device);
    }
    if(getInput(2)) {
        getInput(2)->setBackend(name, device);
    }
    if(getInput(3)) {
        getInput(3)->setBackend(name, device);
    }
}